/*                 
 *      FZ-1 toolkit
 *      (c) 2000 by Rainer Buchty (rainer@buchty.net)
 *            
 *      floppy raw routines (system dependant)
 */

#include"floppy.h"
#include"errno.h"

static char *fdd[]={
	"/dev/fd0",
	"/dev/fd1"
};

typedef struct format_map {
	unsigned char c;
	unsigned char h;
	unsigned char s;
	unsigned char n;
} format_map_t;

char track_buffer[24*512];

void hexdump(char *buf, int size) {
int i,j;
	for(i=0;i<size;i+=16) {
		printf("%04x: ",i);
		for(j=0;j<16;j++) printf("%02x ",buf[i+j]&0xff);
		for(j=0;j<16;j++) if (buf[i+j]<0x20) printf("."); else printf("%c",buf[i+j]);
		printf("\n");
	}
}

int disk_open(int drive) {
int fd;
	if ( (fd=open(fdd[(int)drive],O_ACCMODE|O_NDELAY))<0) {
		printf("Can't access drive %s.\n",fdd[(int)drive]);
		perror("disk_access");
		return -1;
	}
	return fd;
}

int post_rc(int fd, struct floppy_raw_cmd *rc, int rsize) {
	if (ioctl(fd,FDRAWCMD,rc)>=0) {
		if (rc->reply_count<rsize) {
			fprintf(stderr,"Short reply from FDC\n");
			return -1;
		}

		if (rc->flags&0x10000) {
			fprintf(stderr,"FDC returns error.\n");
			return -1;
		} else if (rc->flags&0x20000) {
			fprintf(stderr,"FDC reset / timeout.\n");
			return -1;
		}

		return 0;
	}
	
	switch(errno) {
		case EBUSY:
			fprintf(stderr,"FDC busy, sleeping for a second.\n");
			sleep(1);
			return 1;
		case EIO:
			fprintf(stderr,"resetting controller\n");
			if(ioctl(fd,FDRESET,2)<0) {
				perror("reset");
				return -1;
			}
			return 1;
		default:
			perror("disk access");
			return -1;
	}
}
		

int disk_access(int drive, struct floppy_raw_cmd *rc, int rsize) {
int fd, err;
	if ((fd=disk_open(drive))<0)
		return -1;
	
	err=post_rc(fd,rc,rsize);
	close(fd);

	return err;
}


int disk_reset(int drive) {
int fd,err;
	if ((fd=disk_open(drive))<0)
		return -1;

	err=ioctl(fd,FDRAWCMD,FD_RESET_ALWAYS);

	close(fd);
	return err;
}
	
int parse_error(char st0, char st1, char st2) {
int err=0;
                
	switch (st0&0xc0) {
		case 1:	printf("Disk command terminated abnormally.\n");
			err=-1;
		case 2:	printf("Invalid command.\n");
			err=-1;
		case 3: printf("Abnormal Termination by polling.\n");
			err=-1;
	}
	if (st0&0x10) {
		printf("Panic: Unable to find Track 0.\n");
		err=-1;
	}
/*
	for(i=0;i<8;i++) {
		if (st1&(1<<i)) {
			if (strcmp(st1_err[i],"")!=0) {
				printf("%s.\n",st1_err[i]);
				err=-1;
			}
		}
		if (st2&(1<<i)) {
			if (strcmp(st2_err[i],"")!=0) {
				printf("%s.\n",st2_err[i]);
				err=-1;
			}
		}
	}	
*/
	return err;
}

void init_rc(struct floppy_raw_cmd *rc) {
	rc->flags=0;
	rc->length=0;
	rc->data=NULL;
	rc->rate=0;
	rc->track=0;
}

int read_id(int drive) {
struct floppy_raw_cmd rc;
	init_rc(&rc);

	rc.flags=FD_RAW_INTR;
	rc.cmd_count=2;
	rc.cmd[0]=FD_READID;
	rc.cmd[1]=drive;

	return disk_access(drive,&rc,7);
}	

void seek(int fd, int drive, int c, int h) {	
struct floppy_raw_cmd rc;

	init_rc(&rc);
	rc.flags=FD_RAW_INTR;
	rc.cmd[0]=FD_SEEK;
	rc.cmd[1]=(h<<2)|(drive&3);
	rc.cmd[2]=c;
	rc.cmd_count=3;

	post_rc(fd,&rc,0);
}


int init_disk(int drive) {
struct floppy_raw_cmd rc;
	disk_reset(drive);

	init_rc(&rc);
	rc.flags=FD_RAW_INTR;
	rc.cmd_count=2;
	rc.cmd[0]=FD_RECALIBRATE;
	rc.cmd[1]=drive;

	return read_id(drive);
}

int checkwp(int drive) {
struct floppy_raw_cmd rc;
	init_disk(drive);
	
	init_rc(&rc);
	rc.cmd[0]=FD_GETSTATUS;
	rc.cmd[1]=drive&0x01;
	rc.cmd_count=2;
	rc.flags=0;
	disk_access(drive,&rc,1);

	if (rc.reply[0]&0x40)
		return -1;
	else
		return 0;
}

int track_read(int fd, int drive, int c, int h, int s,int len, char *data) {
struct floppy_raw_cmd rc;

	seek(fd,drive,c,h);

	rc.cmd[0]=FD_READ;
	rc.cmd[1]=(h<<2)|(drive&3);
	rc.cmd[2]=c;
	rc.cmd[3]=h;
	rc.cmd[4]=s;
	rc.cmd[5]=3;
	rc.cmd[6]=0;		// EOT
	rc.cmd[7]=0x35;		// GPL
	rc.cmd[8]=0xff;		// DTL
	rc.cmd_count=9;
	rc.flags=FD_RAW_READ|FD_RAW_INTR;
	rc.data=data;
	rc.rate=0;
	rc.track=c;
	rc.length=len*1024;
		
	post_rc(fd,&rc,7);

	return parse_error(rc.reply[0],rc.reply[1],rc.reply[2]);
}

int m_sector_read(int drive, int ss, int es, char *data) {
unsigned int c,h,s,len,ec,eh,xs;
int fd;
unsigned char *xdata;

	if ((fd=disk_open(drive))<0)
		return -1;
		
	ec=es>>4;
	eh=(es>>3)&0x01;

	xdata=(unsigned char*)malloc(8192);
	for(s=ss;s<=es;) {
		c=s>>4;
		h=(s>>3)&0x01;
		xs=1+(s&0x07);
		
		if ( (c==ec)&&(h==eh) ) {
			if (s!=ss) 
				len=1+(es&0x07);
			else
				len=1+es-ss;
		} else
			len=9-xs;

			
		if (track_read(fd,drive,c,h,xs,len,xdata)<0) {
			free(xdata);
			close(fd);
			return -1;
		}		
		memcpy(data+(s-ss)*1024,xdata,len*1024);
		s+=len;
	}

	free(xdata);
	
	close(fd);
	return 0;
}
 

int m_sector_write(int drive, int ss, int es, char *data) {
int s;

	for(s=ss;s<=es;s++) 
		if (sector_write(0,s,data+1024*(s-ss))<0) return -1;

	return 0;
}

int sector_read(int drive, int s, char *data) {
struct floppy_raw_cmd rc;
int fc,fs,fh,fd;

	if ((fd=disk_open(drive))<0)
		return -1;
		
	fc=s>>4;
	fs=1+(s&0x07);
	fh=(s>>3)&0x01;

	seek(fd,drive,fc,fh);

	rc.cmd[0]=FD_READ;
	rc.cmd[1]=(fh<<2)|(drive&3);
	rc.cmd[2]=fc;
	rc.cmd[3]=fh;
	rc.cmd[4]=fs;
	rc.cmd[5]=3;
	rc.cmd[6]=0;		// EOT
	rc.cmd[7]=0x35;		// GPL
	rc.cmd[8]=0xff;		// DTL
	rc.cmd_count=9;
	rc.flags=FD_RAW_READ|FD_RAW_INTR;
	rc.data=data;
	rc.rate=0;
	rc.track=fc;
	rc.length=1024;
		
	post_rc(fd,&rc,7);
	close(fd);
		
	return parse_error(rc.reply[0],rc.reply[1],rc.reply[2]);
}

int sector_write(int drive, int s, char *data) {
struct floppy_raw_cmd rc;
int fc,fs,fh;

	fc=s>>4;
	fs=1+(s&0x07);
	fh=(s>>3)&0x01;

	seek(drive,drive,fc,fh);

	rc.cmd[0]=FD_WRITE;
	rc.cmd[1]=(fh<<2)|(drive&3);
	rc.cmd[2]=fc;
	rc.cmd[3]=fh;
	rc.cmd[4]=fs;
	rc.cmd[5]=3;
	rc.cmd[6]=0;		// EOT
	rc.cmd[7]=0x35;		// GPL
	rc.cmd[8]=0xff;		// DTL
	rc.cmd_count=9;
	rc.flags=FD_RAW_WRITE|FD_RAW_INTR;
	rc.data=data;
	rc.rate=0;
	rc.track=fc;
	rc.length=1024;
		
	disk_access(drive,&rc,7);
	
	return parse_error(rc.reply[0],rc.reply[1],rc.reply[2]);
}

int format_track(int drive, int c, int h) {
struct floppy_raw_cmd rc;
format_map_t *data;
int i;
char sector[]={1,4,7,2,5,8,3,6};	// interleave 3:1

	data = (format_map_t *) track_buffer;

/*	for(i=0;i<17;++i) {
		data[i].c=c;
		data[i].h=h;
		data[i].s=128+i;
		data[i].n=7;
	}
*/	
	for(i=0;i<8;++i) {
		data[i].c=c;
		data[i].h=h;
		data[i].s=sector[i]-1;
		data[i].n=3;
	}
		
	rc.cmd[0]=FD_FORMAT;
	rc.cmd[1]=(h<<2)|(drive&3);
	rc.cmd[2]=3;		// sector size code
	rc.cmd[3]=8;		// sectors per cylinder
	rc.cmd[4]=0x74;		// gap length
	rc.cmd[5]=0x5a;		// fill byte

	rc.cmd_count=6;
	rc.flags=FD_RAW_WRITE|FD_RAW_INTR|FD_RAW_NEED_SEEK|FD_RAW_NEED_DISK;
	rc.data=data;
	rc.rate=0;
	rc.track=c;
	rc.length=8*sizeof(format_map_t);
		
	disk_access(drive,&rc,7);

	return parse_error(rc.reply[0],rc.reply[1],rc.reply[2]);
}